/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - mib2plugin.c */
/* rdk@krupczak.org */

/* cross-platform definitions for mib-2 plugins for various
   supported platforms (e.g. Linux, Solaris, NT)
*/

/* ************************************************************* */

#ifdef WIN32
#include <windows.h>
#include <IPExport.h>
#endif

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#ifndef WIN32
#include <unistd.h>
#include <strings.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#endif

#include <string.h>
#include <fcntl.h>
#include <stdlib.h>

#include "xmp.h"
#include "xmp_vars.h"
#include "xmp_session.h"
#include "mib2plugin.h"
#include "pluginexport.h"
#include "xmp_events.h"
#include "korgdebug.h"

/* ************************************************************* */

static float mib2PluginVersion = 0.9603;
char *versionStr = "KorgVersion 0.9603";

static xmpErrorStatus_t var_mib2Plugin();

static xmpErrorStatus_t var_ip();
static xmpErrorStatus_t var_icmp();
static xmpErrorStatus_t var_udp();
static xmpErrorStatus_t var_udpTable();
static xmpErrorStatus_t var_tcp();
static xmpErrorStatus_t var_tcpTable();
static xmpErrorStatus_t var_ifTable();
static xmpErrorStatus_t var_addrTable();
static xmpErrorStatus_t var_routeTable();
static xmpErrorStatus_t var_arpTable();

extern ipStats_t ipstats;
extern tcpStats_t tcpstats;
extern icmpStats_t icmpstats;
extern udpStats_t udpstats;

extern pthread_mutex_t mib2_lock;

extern mib2IfEntry_t *ifTable;
extern int numIfs;

extern mib2TcpConn_t *tcpTable;
extern int numTcpConnections;

extern mib2UdpEndpt_t *udpTable;
extern int numUdps;

extern arpEntry_t *arpTable;
extern int numArps;

extern ipRoute_t *routeTable;
extern int numRoutes;

extern ipAddrEntry_t *addrTable;
extern int numAddrs;

extern int debugLevel;
int mib2ThreadKill;

static xmpVar_t mib2Vars[] = {

{"mib2", "pluginVersion", floatingPoint, "", 1,0, var_mib2Plugin, NULL },

{"mib2", "ipForwarding", integer, "", 1,0, var_ip, NULL, NULL, NULL },
{"mib2", "ipDefaultTTL", integer, "", 2,0, var_ip, NULL, NULL, NULL },
{"mib2", "ipInReceives", counter, "", 3,0, var_ip, NULL, NULL, NULL },
{"mib2", "ipInHdrErrors", counter, "",4, 0, var_ip, NULL, NULL, NULL },
{"mib2", "ipInAddrErrors", counter, "",5, 0, var_ip, NULL, NULL, NULL },
{"mib2", "ipForwDatagrams", counter, "",6, 0, var_ip, NULL, NULL, NULL },
{"mib2", "ipInUnknownProtos", counter, "",7, 0, var_ip, NULL, NULL, NULL },
{"mib2", "ipInDiscards", counter, "", 8,0, var_ip, NULL, NULL, NULL },
{"mib2", "ipInDelivers", counter, "", 9,0, var_ip, NULL, NULL, NULL },
{"mib2", "ipOutRequests", counter, "",10, 0, var_ip, NULL, NULL, NULL },
{"mib2", "ipOutDiscards", counter, "",11, 0, var_ip, NULL, NULL, NULL },
{"mib2", "ipOutNoRoutes", counter, "",12, 0, var_ip, NULL, NULL, NULL },
{"mib2", "ipReasmTimeout", counter, "",13, 0, var_ip, NULL, NULL, NULL },
{"mib2", "ipReasmReqds", counter, "", 14,0, var_ip, NULL, NULL, NULL },
{"mib2", "ipReasmOKs", counter, "", 15,0, var_ip, NULL, NULL, NULL },
{"mib2", "ipReasmFails", counter, "", 16, 0, var_ip, NULL, NULL, NULL },
{"mib2", "ipFragOKs", counter, "", 17,0, var_ip, NULL, NULL, NULL },
{"mib2", "ipFragFails", counter, "", 18,0, var_ip, NULL, NULL, NULL },
{"mib2", "ipFragCreates", counter, "", 19,0, var_ip, NULL, NULL, NULL },
{"mib2", "ipRoutingDiscards", counter, "", 20, 0, var_ip, NULL, NULL, NULL },

{"mib2", "icmpInMsgs", counter, "", 21, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpInErrors", counter, "", 22, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpInDestUnreachs", counter, "", 23, 0, var_icmp,NULL,NULL,NULL },
{"mib2", "icmpInTimeExcds", counter, "", 24, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpInParmProbs", counter, "", 25, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpInSrcQuenchs", counter, "", 26, 0, var_icmp,NULL,NULL, NULL },
{"mib2", "icmpInRedirects", counter, "", 27, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpInEchos", counter, "", 28, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpInTimestamps", counter, "", 29, 0, var_icmp,NULL,NULL, NULL },
{"mib2", "icmpInTimestampReps", counter, "", 30, 0, var_icmp,NULL,NULL,NULL},
{"mib2", "icmpInAddrMasks", counter, "", 31, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpInAddrMaskReps", counter, "", 32, 0, var_icmp,NULL,NULL,NULL },
{"mib2", "icmpOutMsgs", counter, "", 33, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpOutErrors", counter, "", 34, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpOutDestUnreachs", counter, "", 35, 0, var_icmp,NULL,NULL,NULL},
{"mib2", "icmpOutTimeExcds", counter, "", 36, 0, var_icmp, NULL,NULL,NULL },
{"mib2", "icmpOutParmProbs", counter, "", 37, 0, var_icmp,NULL,NULL, NULL },
{"mib2", "icmpOutSrcQuenchs", counter, "", 38,0, var_icmp,NULL,NULL,NULL },
{"mib2", "icmpOutRedirects", counter, "", 39, 0, var_icmp,NULL, NULL, NULL },
{"mib2", "icmpOutEchos", counter, "", 40, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpOutEchoReps", counter, "", 41, 0, var_icmp, NULL, NULL, NULL },
{"mib2", "icmpOutTimestamps", counter, "", 42, 0, var_icmp,NULL,NULL,NULL },
{"mib2", "icmpOutTimestampReps", counter, "",43,0,var_icmp,NULL,NULL,NULL },
{"mib2", "icmpOutAddrMasks", counter, "", 44, 0,var_icmp,NULL, NULL, NULL },
{"mib2", "icmpOutAddrMaskReps", counter, "", 45,0, var_icmp,NULL,NULL,NULL},

{"mib2", "udpInDatagrams", counter, "", 46,0, var_udp, NULL, NULL, NULL },
{"mib2", "udpNoPorts", counter, "", 47,0, var_udp, NULL, NULL, NULL },
{"mib2", "udpInErrors", counter, "", 48,0, var_udp, NULL, NULL, NULL },
{"mib2", "udpOutDatagrams", counter, "", 49,0, var_udp, NULL, NULL, NULL },

{"mib2", "tcpRtoAlgorithm", integer, "", 50,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpRtoMin", integer, "", 51,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpRtoMax", integer, "", 52,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpMaxConn", integer, "", 53,0, var_tcp, NULL, NULL, NULL },

{"mib2", "tcpActiveOpens", counter, "", 54,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpPassiveOpens", counter, "", 55,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpAttemptFails", counter, "", 56,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpEstabResets", counter, "", 57,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpCurrEstab", gauge, "", 58,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpInSegs", counter, "", 59,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpOutSegs", counter, "", 60,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpRetransSegs", counter, "", 61,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpInErrs", counter, "", 62,0, var_tcp, NULL, NULL, NULL },
{"mib2", "tcpOutRsts", counter, "", 63,0, var_tcp, NULL, NULL, NULL },

/* udpTable and columns; only support select */
{"mib2", "udpTable", table, "", 64, 0, NULL, NULL, var_udpTable, },
{"mib2", "udpLocalAddress", ipV4Address, "udpTable", 1, 0 },
{"mib2", "udpLocalPort", integer, "udpTable", 2, 0 },

// extensions to rfc1213
{"mib2", "udpEndptState", integer, "udpTable", 3, 0 },
{"mib2", "udpRemAddress",ipV4Address, "udpTable", 4, 0 },
{"mib2", "udpRemPort",integer, "udpTable", 5, 0 },

/* tcpTable and columns; only support select */
{"mib2","tcpConnTable", table, "", 65, 0, NULL, NULL, var_tcpTable, }, 
{"mib2","tcpConnState", integer, "tcpConnTable", 1, 0},
{"mib2","tcpConnLocalAddress", ipV4Address, "tcpConnTable", 2, 0},
{"mib2","tcpConnLocalPort", integer, "tcpConnTable", 3, 0},
{"mib2","tcpConnRemAddress", ipV4Address, "tcpConnTable", 4, 0},
{"mib2","tcpConnRemPort", integer, "tcpConnTable", 5, 0},

 /* ifTable and columns; we only support select ifTable for now */
{"mib2","ifTable", table, "", 66, 0, NULL, NULL, var_ifTable, },
{"mib2","ifDescr",displayString, "ifTable", 1, 0},
{"mib2","ifType", integer, "ifTable", 2, 0},
{"mib2","ifMtu",  integer, "ifTable", 3, 0},
{"mib2","ifSpeed",gauge, "ifTable", 4, 0}, 
{"mib2","ifPhysAddress",physAddress, "ifTable", 5, 0},
{"mib2","ifAdminStatus",integer, "ifTable", 6, 0},
{"mib2","ifOperStatus",integer, "ifTable", 7, 0},
{"mib2","ifInOctets",counter, "ifTable", 8, 0},
{"mib2","ifInUcastPkts",counter, "ifTable", 9, 0},
{"mib2","ifInNucastPkts",counter, "ifTable", 10, 0},
{"mib2","ifInDiscards",counter, "ifTable", 11, 0},
{"mib2","ifInErrors",counter, "ifTable", 12, 0},
{"mib2","ifInUnknownProtos",counter,"ifTable", 13, 0},
{"mib2","ifOutOctets", counter, "ifTable", 14, 0},
{"mib2","ifOutUcastPkts", counter , "ifTable", 15, 0},
{"mib2","ifOutNUcastPkts", counter, "ifTable", 16, 0},
{"mib2","ifOutDiscards", counter, "ifTable", 17, 0},
{"mib2","ifOutErrors", counter, "ifTable", 18, 0},
{"mib2","ifOutQLen", gauge, "ifTable", 19, 0},
{"mib2","ifIndex", integer, "ifTable", 20, 0},

/* ipAddrTable and columns; only support select */

{"mib2", "ipAddrTable", table, "", 67, 0, NULL, NULL, var_addrTable, },
{"mib2", "ipAdEntAddr", ipV4Address, "ipAddrTable",     1, 0},
{"mib2", "ipAdEntIfname", displayString, "ipAddrTable", 2, 0},
{"mib2", "ipAdEntNetMask", ipV4Address, "ipAddrTable",  3, 0},
{"mib2", "ipAdEntBcastAddr", integer, "ipAddrTable",    4, 0},
{"mib2", "ipAdEntReasmMaxSize", integer, "ipAddrTable", 5, 0},

/* route table and columns */

{"mib2", "ipRouteTable", table, "", 68, 0, NULL, NULL, var_routeTable, },
{"mib2", "ipRouteDest", ipV4Address, "ipRouteTable",     1, 0},
{"mib2", "ipRouteIfname", displayString, "ipRouteTable", 2, 0},
{"mib2", "ipRouteMetric", integer, "ipRouteTable",       3, 0},
{"mib2", "ipRouteNextHop", ipV4Address, "ipRouteTable",  4, 0},
{"mib2", "ipRouteType", integer, "ipRouteTable",         5, 0},
{"mib2", "ipRouteProto", integer, "ipRouteTable",        6, 0},
{"mib2", "ipRouteAge", integer, "ipRouteTable",          7, 0},
{"mib2", "ipRouteMask", ipV4Address, "ipRouteTable",     8, 0},

/* arp/ipNetToMedia table and columns */
{"mib2", "ipNetToMediaTable", table, "", 69, 0, NULL, NULL, var_arpTable,},
{"mib2", "ipNetToMediaIfname", displayString, "ipNetToMediaTable", 1, 0},
{"mib2", "ipNetToMediaPhysAddress", physAddress, "ipNetToMediaTable", 2, 0},
{"mib2", "ipNetToMediaNetAddress", ipV4Address, "ipNetToMediaTable", 3, 0},
{"mib2", "ipNetToMediaType", integer, "ipNetToMediaTable", 4, 0},

};

static int numMib2Vars = sizeof(mib2Vars)/sizeof(xmpVar_t);

/* ************************************************************* */
/* convert from in6_addr to our internal format for an IPv6 address 
   right now, only used on Solaris; if used on other platforms,
   need to verify  

   need to verify byte order XXX for Solarisx86
*/

int convertIPv6Addr(struct in6_addr *addr6, char *converted, int maxsize)
{
     if ((addr6 == NULL) || (converted == NULL)) {
        return -1;
     }
     converted[0] = 0;
     if (maxsize < 48) {
        return -1;
     }

     /* take 4-char chunks at a time */
     sprintf(converted,"%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x:%02x%02x",
             addr6->s6_addr[0],
             addr6->s6_addr[1],
	     addr6->s6_addr[2],
	     addr6->s6_addr[3],
	     addr6->s6_addr[4],
	     addr6->s6_addr[5],
	     addr6->s6_addr[6],
	     addr6->s6_addr[7],
	     addr6->s6_addr[8],
	     addr6->s6_addr[9],
	     addr6->s6_addr[10],
	     addr6->s6_addr[11],
	     addr6->s6_addr[12],
	     addr6->s6_addr[13],
	     addr6->s6_addr[14],
	     addr6->s6_addr[15]);

     return strlen(converted);
}


/* ************************************************************* */

static xmpErrorStatus_t var_icmp(xmpVar_t *vp, xmpVariable_t *v) 
{ 
     int ret;
     
     switch (vp->hint) {
       case 21:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInMsgs);
          v->valueStrLen = ret;
 	  break;
       case 22:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInErrors);
          v->valueStrLen = ret;
 	  break;
       case 23:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInDestUnreachs);
          v->valueStrLen = ret;
 	  break;
       case 24:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInTimeExcds);
          v->valueStrLen = ret;
 	  break;
       case 25:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInParmProbs);
          v->valueStrLen = ret;
 	  break;
       case 26:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInSrcQuenchs);
          v->valueStrLen = ret;
 	  break;
       case 27:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInRedirects);
          v->valueStrLen = ret;
 	  break;
       case 28:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInEchos);
          v->valueStrLen = ret;
 	  break;
       case 29:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInTimestamps);
          v->valueStrLen = ret;
 	  break;
       case 30:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInTimestampReps);
          v->valueStrLen = ret;
 	  break;
       case 31:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInAddrMasks);
          v->valueStrLen = ret;
 	  break;
       case 32:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpInAddrMaskReps);
          v->valueStrLen = ret;
 	  break;
       case 33:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutMsgs);
          v->valueStrLen = ret;
 	  break;
       case 34:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutErrors);
          v->valueStrLen = ret;
 	  break;
       case 35:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutDestUnreachs);
          v->valueStrLen = ret;
 	  break;
       case 36:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutTimeExcds);
          v->valueStrLen = ret;
 	  break;
       case 37:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutParmProbs);
          v->valueStrLen = ret;
 	  break;
       case 38:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutSrcQuenchs);
          v->valueStrLen = ret;
 	  break;
       case 39:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutRedirects);
          v->valueStrLen = ret;
 	  break;
       case 40:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutEchos);
          v->valueStrLen = ret;
 	  break;
       case 41:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutEchoReps);
          v->valueStrLen = ret;
 	  break;
       case 42:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutTimestamps);
          v->valueStrLen = ret;
 	  break;
       case 43:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutTimestampReps);
          v->valueStrLen = ret;
 	  break;
       case 44:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutAddrMasks);
          v->valueStrLen = ret;
 	  break;
       case 45:
	  ret = sprintf(v->valueStr,"%llu",icmpstats.icmpOutAddrMaskReps);
          v->valueStrLen = ret;
 	  break;
       default:
  	  v->valueStrLen = 0;
 	  return noSuchName;
     }

     return noError;
}

/* ************************************************************* */

static xmpErrorStatus_t var_udp(xmpVar_t *vp, xmpVariable_t *v) 
{ 
     int ret;

     switch (vp->hint) {
       case 46:
	  ret = sprintf(v->valueStr,"%llu",udpstats.udpInDatagrams);
          v->valueStrLen = ret;
 	  break;
       case 47:
	  ret = sprintf(v->valueStr,"%llu",udpstats.udpNoPorts);
          v->valueStrLen = ret;
 	  break;
       case 48:
	  ret = sprintf(v->valueStr,"%llu",udpstats.udpInErrors);
          v->valueStrLen = ret;
 	  break;
       case 49:
	  ret = sprintf(v->valueStr,"%llu",udpstats.udpOutDatagrams);
          v->valueStrLen = ret;
 	  break;
       default:
  	  v->valueStrLen = 0;
 	  return noSuchName;
     }

     return noError;
}

/* ************************************************************* */

static xmpErrorStatus_t var_tcp(xmpVar_t *vp, xmpVariable_t *v) 
{ 
     int ret;

     switch (vp->hint) {
       case 50:
	  ret = sprintf(v->valueStr,"%lld",tcpstats.tcpRtoAlgorithm);
          v->valueStrLen = ret;
          break;
       case 51:
	  ret = sprintf(v->valueStr,"%lld",tcpstats.tcpRtoMin);
          v->valueStrLen = ret;
          break;
       case 52:
	  ret = sprintf(v->valueStr,"%lld",tcpstats.tcpRtoMax);
          v->valueStrLen = ret;
          break;
       case 53:
	  ret = sprintf(v->valueStr,"%ld",tcpstats.tcpMaxConn);
          v->valueStrLen = ret;
          break;
       case 54:
	  ret = sprintf(v->valueStr,"%llu",tcpstats.tcpActiveOpens);
          v->valueStrLen = ret;
          break;
       case 55:
	  ret = sprintf(v->valueStr,"%llu",tcpstats.tcpPassiveOpens);
          v->valueStrLen = ret;
          break;
       case 56:
	  ret = sprintf(v->valueStr,"%llu",tcpstats.tcpAttemptFails);
          v->valueStrLen = ret;
          break;
       case 57:
	  ret = sprintf(v->valueStr,"%llu",tcpstats.tcpEstabResets);
          v->valueStrLen = ret;
          break;
       case 58:
	  ret = sprintf(v->valueStr,"%lld",tcpstats.tcpCurrEstab);
          v->valueStrLen = ret;
          break;
       case 59:
	  ret = sprintf(v->valueStr,"%llu",tcpstats.tcpInSegs);
          v->valueStrLen = ret;
          break;
       case 60:
	  ret = sprintf(v->valueStr,"%llu",tcpstats.tcpOutSegs);
          v->valueStrLen = ret;
          break;
       case 61:
	  ret = sprintf(v->valueStr,"%llu",tcpstats.tcpRetransSegs);
          v->valueStrLen = ret;
          break;
       case 62:
	  ret = sprintf(v->valueStr,"%llu",tcpstats.tcpInErrs);
          v->valueStrLen = ret;
          break;
       case 63:
	  ret = sprintf(v->valueStr,"%llu",tcpstats.tcpOutRsts);
          v->valueStrLen = ret;
          break;

       default:
  	  v->valueStrLen = 0;
 	  return noSuchName;
     }

     return noError; 
}

/* ************************************************************* */

static xmpErrorStatus_t var_ip(xmpVar_t *vp, xmpVariable_t *v) 
{ 
     int ret;

     switch (vp->hint) {
       case 1:
	  ret = sprintf(v->valueStr,"%ld",ipstats.ipForwarding);
          v->valueStrLen = ret;
          break;
       case 2:
	  ret = sprintf(v->valueStr,"%ld",ipstats.ipDefaultTTL);
          v->valueStrLen = ret;
          break;
       case 3:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipInReceives);
          v->valueStrLen = ret;
          break;
       case 4:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipInHdrErrors);
          v->valueStrLen = ret;
          break;
       case 5:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipInAddrErrors);
          v->valueStrLen = ret;
          break;
       case 6:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipForwDatagrams);
          v->valueStrLen = ret;
          break;
       case 7:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipInUnknownProtos);
          v->valueStrLen = ret;
          break;
       case 8:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipInDiscards);
          v->valueStrLen = ret;
          break;
       case 9:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipInDelivers);
          v->valueStrLen = ret;
          break;
       case 10:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipOutRequests);
          v->valueStrLen = ret;
          break;
       case 11:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipOutDiscards);
          v->valueStrLen = ret;
          break;
       case 12:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipOutNoRoutes);
          v->valueStrLen = ret;
          break;
       case 13:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipReasmTimeout);
          v->valueStrLen = ret;
          break;
       case 14:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipReasmReqds);
          v->valueStrLen = ret;
          break;
       case 15:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipReasmOKs);
          v->valueStrLen = ret;
          break;
       case 16:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipReasmFails);
          v->valueStrLen = ret;
          break;
       case 17:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipFragOKs);
          v->valueStrLen = ret;
          break;
       case 18:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipFragFails);
          v->valueStrLen = ret;
          break;
       case 19:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipFragCreates);
          v->valueStrLen = ret;
          break;
       case 20:
	  ret = sprintf(v->valueStr,"%llu",ipstats.ipRoutingDiscards);
          v->valueStrLen = ret;
          break;
       default:
  	  v->valueStrLen = 0;
 	  return noSuchName;
     }

     return noError; 
}

/* ************************************************************* */

static xmpErrorStatus_t var_arpTable(char *key, xmpVariable_t *columns, 
                                     xmpPdu_t *out)
{
     int numVariables = 0;
     int numRows = 0;
     xmpVariable_t *varptr, *cptr;
     int i;
     u_int maxrows;
     char rowKey[512], aStr1[32];

     if ((key == NULL) || (columns == NULL) || (out == NULL)) {
        return genError;
     }

     /* 
        right now, keys can be specific value, *, \0
        eventually, we will handle regular expressions . . . XXX
     */

     /* get the PDU ready */
     out->numberVariables = 0;
     out->varList = NULL;
     out->varListTail = NULL;
     maxrows = out->maxRows;
     if (maxrows <= 0)
         maxrows = (u_int)-1;

     /* get mib2 lock */
     pthread_mutex_lock(&mib2_lock);

     /* cycle through arpTable */

     for (i=0; i<numArps; i++) {

         /* construct key for this row */
         xmpConvertIpAddr(arpTable[i].ipv4addr,aStr1,sizeof(aStr1));
         sprintf(rowKey,"%s,%s",arpTable[i].ifname,aStr1);

         if ((strcmp(rowKey,key) != 0) && (key[0] != 0) &&
	     (key[0] != '*')) {
             continue; /* with loop */
         }

         cptr = columns;
         while ((cptr != NULL) && (numRows < maxrows)) {

           /* we dont have to check if cptr->entryPtr valid since 
              we would not be called if there was not an internal
              entry for this column */

	   /* malloc a return value */

           if ((varptr = (xmpVariable_t *)malloc(sizeof(xmpVariable_t)))
               == NULL) {
	       perror("varArpTable: malloc failed");
               out->errorStatus = genError;
               /* release lock */
               pthread_mutex_unlock(&mib2_lock);
               return genError;
           }
           bcopy(cptr,varptr,sizeof(xmpVariable_t));
           strcpy(varptr->varName.key,rowKey);

           switch ( ((xmpVar_t *)cptr->entryPtr)->hint) {
	     case 1: /* ifname */
 	        varptr->varSyntax = displayString;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                            arpTable[i].ifname);
	        break;
             case 2: /* physaddress */
	        varptr->varSyntax = physAddress;
                if (arpTable[i].hwaddrlen != 0) {
                   varptr->valueStrLen = sprintf(varptr->valueStr,
                                              "%02x:%02x:%02x:%02x:%02x:%02x",
                                              arpTable[i].hwaddr[0],
                                              arpTable[i].hwaddr[1],
                                              arpTable[i].hwaddr[2],
                                              arpTable[i].hwaddr[3],
                                              arpTable[i].hwaddr[4],
                                              arpTable[i].hwaddr[5]);
                }
                else {
		   varptr->valueStrLen = 0;
                   varptr->valueStr[0] = 0;
		}
	        break;
             case 3: /* netaddress */
	        varptr->varSyntax = ipV4Address;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",aStr1);
	        break;
             case 4: /* mediatype */
	        varptr->varSyntax = integer;
                varptr->valueStrLen=sprintf(varptr->valueStr,"%d",0); /* XXX */
	        break;

 	     default: 
   	        fprintf(stderr,"varArpTable: invalid column\n");
                /* release lock */
                pthread_mutex_unlock(&mib2_lock);
                out->errorStatus = noSuchName;
                out->errorIndex = numVariables;
                out->numberVariables = numVariables;
                free(varptr);
                return noSuchName;

           } /* switch on column */

           /* glue into variable list */
           numVariables++;
           if (out->varListTail)
	      out->varListTail->next = varptr;
           out->varListTail = varptr;
           if (out->varList == NULL)
  	      out->varList = varptr;           

           cptr = cptr->next;

         } /* columns */
         
         numRows++;

     } /* for each arp entry */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     out->numberVariables = numVariables;
     out->errorStatus = noError;
     if (numRows == out->maxRows)
        out->errorStatus = noErrorMoreData;

     if (debugLevel)
        printf("var_arpTable: %d rows, %d objects\n",numRows,numVariables);

     return noError;
}

/* ************************************************************* */

static xmpErrorStatus_t var_routeTable(char *key, xmpVariable_t *columns, 
                                       xmpPdu_t *out)
{
     int numVariables = 0;
     int numRows = 0;
     xmpVariable_t *varptr, *cptr;
     int i;
     u_int maxrows;
     char rowKey[512], aStr1[32];

     if ((key == NULL) || (columns == NULL) || (out == NULL)) {
        return genError;
     }

     /* 
        right now, keys can be specific value, *, \0
        eventually, we will handle regular expressions . . . XXX
     */

     /* get the PDU ready */
     out->numberVariables = 0;
     out->varList = NULL;
     out->varListTail = NULL;
     maxrows = out->maxRows;
     if (maxrows <= 0)
         maxrows = (u_int)-1;

     /* get mib2 lock */
     pthread_mutex_lock(&mib2_lock);

     /* cycle through routeTable */

     for (i=0; i<numRoutes; i++) {

         /* construct key for this row -- assume ipv4 route for now */
         xmpConvertIpAddr(routeTable[i].ipv4dest,aStr1,sizeof(aStr1));
         sprintf(rowKey,"%s",aStr1);

         if ((strcmp(rowKey,key) != 0) && (key[0] != 0) &&
	     (key[0] != '*')) {
             continue; /* with loop */
         }

         cptr = columns;
         while ((cptr != NULL) && (numRows < maxrows)) {

           /* we dont have to check if cptr->entryPtr valid since 
              we would not be called if there was not an internal
              entry for this column */

	   /* malloc a return value */

           if ((varptr = (xmpVariable_t *)malloc(sizeof(xmpVariable_t)))
               == NULL) {
	       perror("varRouteTable: malloc failed");
               out->errorStatus = genError;
               /* release lock */
               pthread_mutex_unlock(&mib2_lock);
               return genError;
           }
           bcopy(cptr,varptr,sizeof(xmpVariable_t));
           strcpy(varptr->varName.key,rowKey);

           switch ( ((xmpVar_t *)cptr->entryPtr)->hint) {
	     case 1: /* dest */
                xmpConvertIpAddr(routeTable[i].ipv4dest,aStr1,sizeof(aStr1));
	        varptr->varSyntax = ipV4Address;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",aStr1);
                break;
	     case 2: /* ifname */
 	        varptr->varSyntax = displayString;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                              routeTable[i].ifname);
                break;
             case 3: /* metric */
	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              routeTable[i].routeMetric);
                break;
             case 4: /* nexthop */
                xmpConvertIpAddr(routeTable[i].ipv4gway,aStr1,sizeof(aStr1));
	        varptr->varSyntax = ipV4Address;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",aStr1);
                break;
             case 5: /* type */
	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              routeTable[i].routeType);
                break;
             case 6: /* proto */
	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              routeTable[i].routeProto);
                break;
             case 7: /* age */
	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              routeTable[i].routeAge);
                break;
             case 8: /* mask */
                xmpConvertIpAddr(routeTable[i].ipv4mask,aStr1,sizeof(aStr1));
	        varptr->varSyntax = ipV4Address;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",aStr1);
                break;

 	     default: 
   	        fprintf(stderr,"varRouteTable: invalid column\n");
                /* release lock */
                pthread_mutex_unlock(&mib2_lock);
                out->errorStatus = noSuchName;
                out->errorIndex = numVariables;
                out->numberVariables = numVariables;
                return noSuchName;

           } /* switch on column */

           /* glue into variable list */
           numVariables++;
           if (out->varListTail)
	      out->varListTail->next = varptr;
           out->varListTail = varptr;
           if (out->varList == NULL)
  	      out->varList = varptr;           

           cptr = cptr->next;

         } /* columns */
         
         numRows++;

     } /* for each route entry */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     out->numberVariables = numVariables;
     out->errorStatus = noError;
     if (numRows == out->maxRows)
        out->errorStatus = noErrorMoreData;

     if (debugLevel)
        printf("var_routeTable: %d rows, %d objects\n",numRows,numVariables);

     return noError;
}

/* ************************************************************* */

static xmpErrorStatus_t var_addrTable(char *key, xmpVariable_t *columns, 
                                      xmpPdu_t *out)
{
     int numVariables = 0;
     int numRows = 0;
     xmpVariable_t *varptr, *cptr;
     int i;
     u_int maxrows;
     char rowKey[512], aStr1[64];

     if ((key == NULL) || (columns == NULL) || (out == NULL)) {
        return genError;
     }

     /* 
        right now, keys can be specific value, *, \0
        eventually, we will handle regular expressions . . . XXX
     */

     /* get the PDU ready */
     out->numberVariables = 0;
     out->varList = NULL;
     out->varListTail = NULL;
     maxrows = out->maxRows;
     if (maxrows <= 0)
         maxrows = (u_int)-1;

     /* get mib2 lock */
     pthread_mutex_lock(&mib2_lock);

     /* cycle through addrTable */

     for (i=0; i<numAddrs; i++) {

         /* construct key for this row */
         if (addrTable[i].ipVersion == 4) {
            xmpConvertIpAddr(addrTable[i].ipv4addr,aStr1,sizeof(aStr1));
         }
         else {
	    strcpy(aStr1,addrTable[i].ipv6Addr);
         }
         sprintf(rowKey,"%s,%s",addrTable[i].ifname,aStr1);

         if ((strcmp(rowKey,key) != 0) && (key[0] != 0) &&
	     (key[0] != '*')) {
             continue; /* with loop */
         }

         cptr = columns;
         while ((cptr != NULL) && (numRows < maxrows)) {

           /* we dont have to check if cptr->entryPtr valid since 
              we would not be called if there was not an internal
              entry for this column */

	   /* malloc a return value */

           if ((varptr = (xmpVariable_t *)malloc(sizeof(xmpVariable_t)))
               == NULL) {
	       perror("varAddrTable: malloc failed");
               out->errorStatus = genError;
               /* release lock */
               pthread_mutex_unlock(&mib2_lock);
               return genError;
           }
           bcopy(cptr,varptr,sizeof(xmpVariable_t));
           strcpy(varptr->varName.key,rowKey);

           switch ( ((xmpVar_t *)cptr->entryPtr)->hint) {

 	     case 1: /* ipaddr */
	        varptr->varSyntax = ipV4Address;
                if (addrTable[i].ipVersion == 6) {
	 	   varptr->varSyntax = ipV6Address;
		}
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",aStr1);
	        break;

	     case 2: /* ifname */
	        varptr->varSyntax = displayString;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                              addrTable[i].ifname);
	        break;

  	     case 3: /* netmask */
	        varptr->varSyntax = ipV4Address;
                xmpConvertIpAddr(addrTable[i].ipv4mask,aStr1,sizeof(aStr1));
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",aStr1);
                if (addrTable[i].ipVersion == 6) {
	 	   varptr->valueStrLen = sprintf(varptr->valueStr,"0.0.0.0");
                }
	        break;

  	     case 4: /* bcast addr */
	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              addrTable[i].ipv4bcast);
	        break;
   	     case 5: /* reasmmaxsize */
	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              addrTable[i].ipv4ReasmMaxSize);
	        break;

 	     default:
   	        fprintf(stderr,"varAddrTable: invalid column\n");
                /* release lock */
                pthread_mutex_unlock(&mib2_lock);
                out->errorStatus = noSuchName;
                out->errorIndex = numVariables;
                out->numberVariables = numVariables;
                return noSuchName;

           } /* switch on column */

           /* glue into variable list */
           numVariables++;
           if (out->varListTail)
	      out->varListTail->next = varptr;
           out->varListTail = varptr;
           if (out->varList == NULL)
  	      out->varList = varptr;           

           cptr = cptr->next;

         } /* columns */
         
         numRows++;

     } /* for each addr entry */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     out->numberVariables = numVariables;
     out->errorStatus = noError;
     if (numRows == out->maxRows)
        out->errorStatus = noErrorMoreData;

     if (debugLevel)
        printf("var_addrTable: %d rows, %d objects\n",numRows,numVariables);

     return noError;
}

/* ************************************************************* */

static xmpErrorStatus_t var_udpTable(char *key, xmpVariable_t *columns, 
                                    xmpPdu_t *out)
{
     int numVariables = 0;
     int numRows = 0;
     xmpVariable_t *varptr, *cptr;
     int i;
     u_int maxrows;
     char rowKey[512], aStr1[64], aStr2[64];

     if ((key == NULL) || (columns == NULL) || (out == NULL)) {
        return genError;
     }

     /* 
        right now, keys can be specific value, *, \0
        eventually, we will handle regular expressions . . . XXX
     */

     /* get the PDU ready */
     out->numberVariables = 0;
     out->varList = NULL;
     out->varListTail = NULL;
     maxrows = out->maxRows;
     if (maxrows <= 0)
         maxrows = (u_int)-1;

     /* get mib2 lock */
     pthread_mutex_lock(&mib2_lock);

     /* cycle through udpTable */

     for (i=0; i<numUdps; i++) {

         /* construct key for this row */
         if (udpTable[i].ipVersion == 4) {
            xmpConvertIpAddr(udpTable[i].localIp,aStr1,sizeof(aStr1));
            xmpConvertIpAddr(udpTable[i].remIp,aStr2,sizeof(aStr2));
         }
         else {
	    /* IPv6 */
	    strcpy(aStr1,udpTable[i].localIPv6);
	    strcpy(aStr2,udpTable[i].remIPv6);
         }

         sprintf(rowKey,"%s:%d",
		 aStr1,udpTable[i].localPort);

         if ((strcmp(rowKey,key) != 0) && (key[0] != 0) &&
	     (key[0] != '*')) {
             continue; /* with loop */
         }

         cptr = columns;
         while ((cptr != NULL) && (numRows < maxrows)) {

           /* we dont have to check if cptr->entryPtr valid since 
              we would not be called if there was not an internal
              entry for this column */

	   /* malloc a return value */

           if ((varptr = (xmpVariable_t *)malloc(sizeof(xmpVariable_t)))
               == NULL) {
	       perror("varUdpTable: malloc failed");
               out->errorStatus = genError;
               /* release lock */
               pthread_mutex_unlock(&mib2_lock);
               return genError;
           }
           bcopy(cptr,varptr,sizeof(xmpVariable_t));
           strcpy(varptr->varName.key,rowKey);

           switch ( ((xmpVar_t *)cptr->entryPtr)->hint) {

	     case 1: /* local Addr */
 	        varptr->varSyntax = ipV4Address;
                if (udpTable[i].ipVersion == 6)
      	           varptr->varSyntax = ipV6Address;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",aStr1);
	        break;

 	     case 2: /* local port */
 	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              udpTable[i].localPort);
	        break;

  	     case 3:  /* endpoint state -- extension to rfc1213 */
 	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              udpTable[i].state);
	        break;

#ifndef WIN32
	     case 4: /* remote address -- extension to rfc1213 */
 	        varptr->varSyntax = ipV4Address;
                if (udpTable[i].ipVersion == 6)
      	           varptr->varSyntax = ipV6Address;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",aStr2);
	        break;

	     case 5: /* remote port -- extension to rfc1213 */
 	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              udpTable[i].remPort);
	        break;
#endif /* UNIX */

#ifdef WIN32
	     case 4: /* remote address -- extension to rfc1213 */
	     case 5: /* remote port -- extension to rfc1213 */
	        varptr->varSyntax = unsupportedVariable;
                varptr->valueStr[0] = 0;
                varptr->valueStrLen = 0;
                break;
#endif /* WIN32 */

 	     default:           
   	        fprintf(stderr,"varUdpTable: invalid column\n");
                /* release lock */
                pthread_mutex_unlock(&mib2_lock);
                out->errorStatus = noSuchName;
                out->errorIndex = numVariables;
                out->numberVariables = numVariables;
                return noSuchName;

           } /* switch on column */

           /* glue into variable list */
           numVariables++;
           if (out->varListTail)
	      out->varListTail->next = varptr;
           out->varListTail = varptr;
           if (out->varList == NULL)
  	      out->varList = varptr;           

           cptr = cptr->next;

         } /* columns */
         
         numRows++;

     } /* for each udp endpt */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     out->numberVariables = numVariables;
     out->errorStatus = noError;
     if (numRows == out->maxRows)
        out->errorStatus = noErrorMoreData;

     if (debugLevel)
        printf("var_udpTable: %d rows, %d objects\n",numRows,numVariables);

     return noError;
}

/* ************************************************************* */

static xmpErrorStatus_t var_tcpTable(char *key, xmpVariable_t *columns, 
                                    xmpPdu_t *out)
{
     int numVariables = 0;
     int numRows = 0;
     xmpVariable_t *varptr, *cptr;
     int i;
     u_int maxrows;
     char rowKey[512], aStr1[64], aStr2[64];

     if ((key == NULL) || (columns == NULL) || (out == NULL)) {
        return genError;
     }

     /* 
        right now, keys can be specific value, *, \0
        eventually, we will handle regular expressions . . . XXX
     */

     /* get the PDU ready */
     out->numberVariables = 0;
     out->varList = NULL;
     out->varListTail = NULL;
     maxrows = out->maxRows;
     if (maxrows <= 0)
         maxrows = (u_int)-1;

     /* get mib2 lock */
     pthread_mutex_lock(&mib2_lock);

     /* cycle through tcpTable */

     for (i=0; i<numTcpConnections; i++) {

         /* construct key for this row */

         if (tcpTable[i].ipVersion == 4) {
            xmpConvertIpAddr(tcpTable[i].localIp,aStr1,sizeof(aStr1));
            xmpConvertIpAddr(tcpTable[i].remIp,aStr2,sizeof(aStr2));
         }
         else {
	    /* IPv6 */
	    strcpy(aStr1,tcpTable[i].localIPv6);
	    strcpy(aStr2,tcpTable[i].remIPv6);
         }
         sprintf(rowKey,"%s:%d:+:%s:%d",
                 aStr1,tcpTable[i].localPort,aStr2,tcpTable[i].remPort);

         if ((strcmp(rowKey,key) != 0) && (key[0] != 0) &&
	     (key[0] != '*')) {
             continue; /* with loop */
         }

         cptr = columns;
         while ((cptr != NULL) && (numRows < maxrows)) {

           /* we dont have to check if cptr->entryPtr valid since 
              we would not be called if there was not an internal
              entry for this column */

	   /* malloc a return value */

           if ((varptr = (xmpVariable_t *)malloc(sizeof(xmpVariable_t)))
               == NULL) {
	       perror("vartcpTable: malloc failed");
               out->errorStatus = genError;
               /* release lock */
               pthread_mutex_unlock(&mib2_lock);
               return genError;
           }
           bcopy(cptr,varptr,sizeof(xmpVariable_t));
           strcpy(varptr->varName.key,rowKey);

           switch ( ((xmpVar_t *)cptr->entryPtr)->hint) {

 	     case 1: /* state */
 	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              tcpTable[i].state);
	        break;

	     case 2: /* local Addr */
 	        varptr->varSyntax = ipV4Address;
                if (tcpTable[i].ipVersion == 6)
     	           varptr->varSyntax = ipV6Address;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",aStr1);
	        break;

	     case 3: /* local port */
 	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              tcpTable[i].localPort);
	        break;

  	     case 4: /* rem Addr */
 	        varptr->varSyntax = ipV4Address;
                if (tcpTable[i].ipVersion == 6)
     	           varptr->varSyntax = ipV6Address;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",aStr2);
	        break;

  	     case 5: /* rem port */
 	        varptr->varSyntax = integer;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              tcpTable[i].remPort);
	        break;

 	     default:           
   	        fprintf(stderr,"vartcpTable: invalid column\n");
                /* release lock */
                pthread_mutex_unlock(&mib2_lock);
                out->errorStatus = noSuchName;
                out->numberVariables = numVariables;
                out->errorIndex = numVariables;
                return noSuchName;

           } /* switch on column */

           /* glue into variable list */
           numVariables++;
           if (out->varListTail)
	      out->varListTail->next = varptr;
           out->varListTail = varptr;
           if (out->varList == NULL)
  	      out->varList = varptr;           

           cptr = cptr->next;

         } /* columns */
         
         numRows++;

     } /* for each tcp connection */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     out->numberVariables = numVariables;
     out->errorStatus = noError;
     if (numRows == out->maxRows)
        out->errorStatus = noErrorMoreData;

     if (debugLevel)
        printf("var_tcpTable: %d rows, %d objects\n",numRows,numVariables);

     return noError;
}

/* ************************************************************* */

static xmpErrorStatus_t var_ifTable(char *key, xmpVariable_t *columns, 
                                    xmpPdu_t *out)
{
     int numVariables = 0;
     int numRows = 0;
     xmpVariable_t *varptr, *cptr;
     int i;
     u_int maxrows;
     char rowKey[512];
     char rowAlias[512];

     if ((key == NULL) || (columns == NULL) || (out == NULL)) {
        return genError;
     }

     if (debugLevel)
        printf("var_ifTable: starting, maxRows=%d, key='%s'\n",out->maxRows,
               key);

     /* 
        right now, keys can be specific value, *, \0
        eventually, we will handle regular expressions . . . XXX
     */

     /* get the PDU ready */
     out->numberVariables = 0;
     out->varList = NULL;
     out->varListTail = NULL;
     maxrows = out->maxRows;
     if (maxrows <= 0)
         maxrows = (u_int)-1;

     if (debugLevel)
         printf("var_ifTable: maxRows=%d, numIfs=%d, columns=%lx\n",
                maxrows,numIfs,columns);

     /* get mib2 lock */
     pthread_mutex_lock(&mib2_lock);

     /* cycle through ifTable */
     for (i=0; i<numIfs; i++) {

         /* construct key for this row */
         strncpy(rowKey,ifTable[i].ifname,255); rowKey[255] = 0;
         strncpy(rowAlias,ifTable[i].alias,255); rowAlias[255] = 0;

         /* if row is a match, cycle through desired columns, else
            continue searching rows */

         if ((strcmp(rowKey,key) != 0) && (key[0] != 0) &&
             (strcmp(rowAlias,key) != 0) && (key[0] != '*')) {
  	    DBG2("var_ifTable: keys '%s' '%s' dont match\n",rowKey,key);
	    continue; /* with for loop */
         }

         cptr = columns;
         while ((cptr != NULL) && (numRows < maxrows)) {

           /* we dont have to check if cptr->entryPtr valid since 
              we would not be called if there was not an internal
              entry for this column */

	   /* malloc a return value */

           if ((varptr =(xmpVariable_t *)malloc(sizeof(xmpVariable_t)))
                 == NULL) {
	      perror("varIfTable: malloc failed");
              out->errorStatus = genError;
              /* release lock */
              pthread_mutex_unlock(&mib2_lock);
              return genError;
           }
           bcopy(cptr,varptr,sizeof(xmpVariable_t));
           strcpy(varptr->varName.key,rowKey);
           if (debugLevel)
              printf("var_IfTable: malloc'd variable for %d\n",
                    ((xmpVar_t *)(cptr->entryPtr))->hint);

           switch ( ((xmpVar_t *)cptr->entryPtr)->hint) {

 	     case 1: /* ifDescr */
	        varptr->varSyntax = displayString;
		varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                              ifTable[i].ifname);
		break;
	     case 2: /* ifType */
 	        varptr->varSyntax = integer;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              ifTable[i].ifType);
		break;
	     case 3: /* ifMtu */
	        varptr->varSyntax = integer;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              ifTable[i].mtu);
		break;
	     case 4: /* ifSpeed */
	        varptr->varSyntax = gauge;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              ifTable[i].ifSpeed);
		break;
	     case 5: /* ifPhyAddress */
	        varptr->varSyntax = physAddress;
                if (ifTable[i].hwaddrlen != 0) {
                   varptr->valueStrLen = sprintf(varptr->valueStr,
                                              "%02x:%02x:%02x:%02x:%02x:%02x",
                                              ifTable[i].hwaddr[0],
                                              ifTable[i].hwaddr[1],
                                              ifTable[i].hwaddr[2],
                                              ifTable[i].hwaddr[3],
                                              ifTable[i].hwaddr[4],
                                              ifTable[i].hwaddr[5]);
                }
                else {
		   varptr->valueStrLen = 0;
                   varptr->valueStr[0] = 0;
		}
		break;
	     case 6: /* ifAdminStatus */
	        varptr->varSyntax = integer;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              ifTable[i].ifAdminStatus);
		break;
	     case 7: /* ifOperStatus */
	        varptr->varSyntax = integer;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              ifTable[i].ifOperStatus);
		break;
	     case 8: /* ifInOctets */
	        varptr->varSyntax = counter;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifInOctets);
		break;
	     case 9: /* ifInUcastPkts */
                varptr->varSyntax = counter;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifInUcastPkts);
		break;
	     case 10: /* ifNUcastPkts */
                varptr->varSyntax = counter;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifInNUcastPkts);
		break;
	     case 11: /* ifInDiscards */
                varptr->varSyntax = counter;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifInDiscards);
		break;
	     case 12: /* ifInErrors */
                varptr->varSyntax = counter;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifInErrors);
		break;
	     case 13: /* ifInUnknownProtos */
                varptr->varSyntax = counter;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifInUnknownProtos);
		break;
	     case 14: /* ifOutOctets */
                varptr->varSyntax = counter;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifOutOctets);
		break;
	     case 15: /* ifOutUcastPkts */
                varptr->varSyntax = counter;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifOutUcastPkts);
		break;
	     case 16: /* ifOutNUcastPkts */
                varptr->varSyntax = counter;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifOutNUcastPkts);
		break;
	     case 17: /* ifOutDiscards */
                varptr->varSyntax = counter;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifOutDiscards);
		break;
	     case 18: /* ifOutErrors */
                varptr->varSyntax = counter;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%llu",
                                              ifTable[i].ifOutErrors);
		break;
	     case 19: /* ifOutQLen */
 	        varptr->varSyntax = gauge;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              ifTable[i].ifOutQLen);
		break;
	     case 20: /* ifIndex */
	        varptr->varSyntax = integer;
	        varptr->valueStrLen = sprintf(varptr->valueStr,"%d",
                                              ifTable[i].ifIndex);
                break;


             default:
   	        fprintf(stderr,"varIfTable: invalid column\n");
                /* release lock */
                pthread_mutex_unlock(&mib2_lock);
                out->errorStatus = noSuchName;
                out->errorIndex = numVariables;
                out->numberVariables = numVariables;
                return noSuchName;

           } /* switch on cptr->entryPtr->hint  */

           /* glue into variable list */
           numVariables++;
           if (out->varListTail)
	      out->varListTail->next = varptr;
           out->varListTail = varptr;
           if (out->varList == NULL)
  	      out->varList = varptr;           

           cptr = cptr->next;

         } /* columns */

         numRows++;

     } /* for each interface */

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     out->numberVariables = numVariables;
     out->errorStatus = noError;
     if (numRows == out->maxRows)
        out->errorStatus = noErrorMoreData;

     if (debugLevel)
        printf("var_ifTable: %d rows, %d objects\n",numRows,numVariables);

     return noError;
}

/* ************************************************************* */

static xmpErrorStatus_t var_mib2Plugin(xmpVar_t *vp, xmpVariable_t *v)
{
     int ret;

     switch (vp->hint) {
       case 1: /* plugin Version */
	  ret = sprintf(v->valueStr,"%3.4f",mib2PluginVersion); 
          v->valueStrLen = ret;
          break;

       default:
  	  v->valueStrLen = 0;
 	  return noSuchName;
     }

     return noError;
}

/* ************************************************************* */

DLLExport int modinfo(char *modname, int modnamesz, char *descr, int descrsz)
{
     printf("mib2Plugin: modinfo called\n");

     if ((modname == NULL) || (descr == NULL)) {
        return -1;
     }

     if ((modnamesz < 11) || (descrsz < 30)) {
        return -1;
     }

     sprintf(modname,"mib2Plugin");
     sprintf(descr,"mib2Plugin version %3.4f",mib2PluginVersion);
#ifdef LINUX
     sprintf(descr,"mib2Plugin Linux version %3.4f",mib2PluginVersion);
#endif
#ifdef SOLARIS
     sprintf(descr,"mib2Plugin Solaris version %3.4f",mib2PluginVersion);
#endif
#ifdef WIN32
     sprintf(descr,"mib2Plugin Win32 version %3.4f",mib2PluginVersion);
#endif

     return 1;
}

/* ************************************************************* */

DLLExport int finish()
{
     printf("Mib2: finish called\n");

     /* need to kill any threads that we started;
        use pthread_kill() to send signal to thread(s)
        running within me and have them exit */

     /* or, use pthread_cancel() to cancel threads
        that are at various cancel points set by
        threads calling pthread_testcancel() */
         
     /* kill threads by setting variable and have thread watch for it */
     /* XXX need to test for this so that when we return from this 
        entry point, we know thread has died */
     mib2ThreadKill = 1;

     /* unload any MIB objects we may have registered */
     xmpRemoveVarByName("mib2",NULL);

     printf("Mib2 finish returning");

     xmpdLog("Mib2 finish returning");

     return 1;
}

/* ************************************************************* */


/* ************************************************************* */
/* ************************************************************* */
/* ************************************************************* */
/* externally visible functions for invocation by other plugins */

DLLExport int mib2GetNumberUdpEndpoints()
{
     int ret;

     pthread_mutex_lock(&mib2_lock);
     ret = numUdps;
     pthread_mutex_unlock(&mib2_lock);

     return ret;
}

/* ************************************************************* */

DLLExport int mib2GetUdpEndpoints(mib2UdpEndpt_t *copytable, int maxudps)
{
     int ret;

     if (copytable == NULL)
        return -1;
     if (maxudps < 5)
        return -1;

     pthread_mutex_lock(&mib2_lock);

     ret = maxudps;

     if (ret > numUdps)
        ret = numUdps;

     if (ret < numUdps) {
        ; /* caller gets less data than exists */
     }

     bcopy(udpTable,copytable,ret*sizeof(mib2UdpEndpt_t));
     pthread_mutex_unlock(&mib2_lock);

     return ret;
}

/* ************************************************************* */
/* 
   get number of tcp connections currently in cache; caller can
   use this number to pre-allocate space in order to call
   function below to get a copy of the table; relevant structures in
   mib2plugin.h are then part of API
*/

DLLExport int mib2GetNumberTcpConnections()
{
     int ret;

     pthread_mutex_lock(&mib2_lock);
     ret = numTcpConnections;  
     pthread_mutex_unlock(&mib2_lock);

     return ret;
}

/* ************************************************************* */
/* copy up to maxconnections tcp connection entries into space
   provided by caller; return number actually copied */

DLLExport int mib2GetTcpConnections(mib2TcpConn_t *copytable, int maxconnections)
{
     int ret;

     if (copytable == NULL)
        return -1;
     if (maxconnections < 5)
        return -1;

     pthread_mutex_lock(&mib2_lock);

     ret = maxconnections;
     if (ret > numTcpConnections)
        ret = numTcpConnections;

     if (ret < numTcpConnections) {
        ; /* caller gets less data than exists */
     }

     bcopy(tcpTable,copytable,ret*sizeof(mib2TcpConn_t));
     pthread_mutex_unlock(&mib2_lock);

     return ret;
}

/* ************************************************************* */
/* get number of routes */

DLLExport int mib2GetNumberRoutes()
{
     int ret;

     pthread_mutex_lock(&mib2_lock);
     ret = numRoutes;
     pthread_mutex_unlock(&mib2_lock);

     return ret;
}

/* ************************************************************* */

DLLExport int mib2GetRoutes(ipRoute_t *copytable, int maxroutes)
{
     int ret;

     if (copytable == NULL)
        return -1;
     if (maxroutes < 5)
        return -1;

     pthread_mutex_lock(&mib2_lock);

     ret = maxroutes;
     if (ret > numRoutes)
        ret = numRoutes;

     if (ret < numRoutes) {
        ; /* caller gets less data than exists */
     }

     bcopy(routeTable,copytable,ret*sizeof(ipRoute_t));
     pthread_mutex_unlock(&mib2_lock);

     return ret;
}

/* ************************************************************* */

DLLExport int mib2GetNumberArps()
{
     int ret;

     pthread_mutex_lock(&mib2_lock);
     ret = numArps;
     pthread_mutex_unlock(&mib2_lock);

     return ret;
}

DLLExport int mib2GetArpTable(arpEntry_t *copytable, int maxarps)
{
     int ret;

     if (copytable == NULL)
        return -1;
     if (maxarps < 5)
        return -1;

     pthread_mutex_lock(&mib2_lock);

     ret = maxarps;
     if (ret > numArps)
        ret = numArps;

     if (ret < numArps) {
        ; /* caller gets less data than exists */
     }

     bcopy(arpTable,copytable,ret*sizeof(arpEntry_t));
     pthread_mutex_unlock(&mib2_lock);

     return ret;
}

/* ************************************************************* */
/* given an address in string/ascii format, is it local? */

DLLExport int mib2isLocalAddressStr(char *addrstr)
{
     u_int addr;
     int ret,i;

     if (addrstr == NULL)
        return 0;

     /* if its a name, we need to convert to ip address XXX */

     addr = (u_int)inet_addr(addrstr);

     /* get mib2 lock */
     pthread_mutex_lock(&mib2_lock);

     ret = 0;
     for (i=0; i<numAddrs; i++) {
         if (addrTable[i].ipv4addr == addr) {
	     ret = 1;
             break; /* from for loop */
         }
     }

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     return ret;
}

/* ************************************************************* */
/* updated for ipv6 and ipv4 addresses and to return 0 if
   no addresses were fetched */

/* make sure to *always* add a comma after ipddress,
   even the last one, so that our matching function can avoid
   10.0.0.1 matching 10.0.0.10 !!!! XXX */

DLLExport int mib2GetAddressList(char *addrlist, int maxlen)
{
     int len,i;
     char addrstr[64];

     if ((addrlist == NULL) || (maxlen < 10))
        return -1;

     /* get mib2 lock */
     pthread_mutex_lock(&mib2_lock);

     /* if there are no addresses yet to be returned, return 0 */
     if (numAddrs < 1) {
        fprintf(stderr,"mib2GetAddressList: no addresses yet!!!\n");
        pthread_mutex_unlock(&mib2_lock);       
        return 0;
     }

     len = 0;
     addrlist[0] = 0;
     i = 0;
     addrstr[0] = 0;

     while ((i < numAddrs) && (len < maxlen)) {

       /* should check if its ipv6 addr XXX */
       if (addrTable[i].ipVersion == 4)
          xmpConvertIpAddr(addrTable[i].ipv4addr,addrstr,63);
       else
 	  strcpy(addrstr,addrTable[i].ipv6Addr);
 
       strncat(addrlist,addrstr,maxlen-len);
       len += strlen(addrstr);
       strncat(addrlist,",",maxlen-len);
       len += 1;
       i++;

     }

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);

     return 1;
}

/* ************************************************************* */
/* given an IPv4 addr, is it local? */

DLLExport int mib2isLocalAddress(u_int addr)
{
     int ret,i;

     /* get mib2 lock */
     pthread_mutex_lock(&mib2_lock);

     ret = 0;
     for (i=0; i<numAddrs; i++) {
       if (addrTable[i].ipv4addr == addr) {
	   ret = 1;
           break; /* from for loop */
       }
     }

     /* release lock */
     pthread_mutex_unlock(&mib2_lock);
 
     return ret;
}

/* ************************************************************* */

#ifdef WIN32

/* a DLL entry point called upon load/free; we dont normally
   need to do anything but some plugins may need this
   level of initialization/uninitialization */

BOOL WINAPI DLLMain(HINSTANCE hDll, DWORD Reason, LPVOID Reserved)
{
     return TRUE;
}

#endif /* WIN32 */

/* ************************************************************* */
/* plugin init, all platforms; rest of plugin entry points are
   in platform-specific modules 
*/

DLLExport int init()
{
     int i;
     char version[64];
     float v;

     printf("mib2plugin: init called\n");
     DBG1("mib2plugin: mib2GetNumberTcpConnections=%lx\n",
          mib2GetNumberTcpConnections);

     bzero(&ipstats,sizeof(ipstats));
     bzero(&icmpstats,sizeof(icmpstats));
     bzero(&tcpstats,sizeof(tcpstats));
     bzero(&udpstats,sizeof(udpstats));

     debugLevel = xmpGetAgentDebugLevel();

     /* add managed objects */
     for (i=0; i<numMib2Vars; i++) {
         xmpAddVar(&mib2Vars[i]);
     }

     v = xmpFindBinaryVersion("mib2-linux.so",NULL);
     sprintf(version,"%3.4f",v);
     xmpAddModule("mib2-linux.so",version,"mib2",NULL,2);

     v = xmpFindBinaryVersion("mib2-solaris.so",NULL);
     sprintf(version,"%3.4f",v);
     xmpAddModule("mib2-solaris.so",version,"mib2",NULL,2);

     v = xmpFindBinaryVersion("mib2-solx86.so",NULL);
     sprintf(version,"%3.4f",v);
     xmpAddModule("mib2-solx86.so",version,"mib2",NULL,2);

     v = xmpFindBinaryVersion("mib2-win32.dll",NULL);
     sprintf(version,"%3.4f",v);
     xmpAddModule("mib2-win32.dll",version,"mib2",NULL,2);

     /* register which types of traps we wish to see XXX */

     return 1;
}

/* ************************************************************* */

DLLExport int mib2TrapHandler(xmpEventEntry_t *e)
{
     DBG2("mib2TrapHandler: event '%s' from %s\n",e->eventType,e->eventSource);

     if (e->eventSource[0] == 0)
        return 0;

     return 1;
}

/* ************************************************************* */
/* transition of ifOperStatus from down state to some other state */

int sendLinkUpTrap(int ifIndex, int ifAdminStatus, int ifOperStatus, char *ifName)
{
     xmpPdu_t *t;
     xmpVariable_t *v1,*v2,*v3,*v4,*v5;
     int ret;

     /* each of the variables needs to be malloc'd separately since
        they are each freed separately XXXX */

     fprintf(stderr,"sendLinkUpTrap: %d %d %d\n",ifIndex,ifAdminStatus,
             ifOperStatus);

     /* alloc a PDU and get it set up */
     if ((t = (xmpPdu_t *)malloc(sizeof(xmpPdu_t))) == NULL) {
        return -1;
     }
     if ((v1 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return -1;
     }
     if ((v2 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return -1;
     }
     if ((v3 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return -1;
     }
     if ((v4 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return -1;
     }
     if ((v5 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return -1;
     }

     bzero((void *)t,sizeof(xmpPdu_t));
     bzero((void *)v1,sizeof(xmpVariable_t));
     bzero((void *)v2,sizeof(xmpVariable_t));
     bzero((void *)v3,sizeof(xmpVariable_t));
     bzero((void *)v4,sizeof(xmpVariable_t));
     bzero((void *)v5,sizeof(xmpVariable_t));
     t->msgType = Trap;
     t->messageID = (int64_t)time(NULL);
     t->errorStatus = noError;
     t->errorIndex = 0;
     t->numberVariables = 5;
     t->varList = v1;
     t->varListTail = v5;

     sprintf(v1->varName.mibName,"core");
     sprintf(v1->varName.objName,"trapType");
     v1->varSyntax = displayString;
     v1->valueStrLen = sprintf(v1->valueStr,"mib2.ifLinkUp");
     v1->next = v2;

     sprintf(v2->varName.mibName,"core");
     sprintf(v2->varName.objName,"sysObjectID");
     v2->varSyntax = displayString;
     v2->valueStrLen = sprintf(v2->valueStr,KORGOBJID);
     v2->next = v3;

     sprintf(v3->varName.mibName,"mib2");
     sprintf(v3->varName.objName,"ifIndex");
     sprintf(v3->varName.key,"%s",ifName);
     v3->varSyntax = integer;
     v3->valueStrLen = sprintf(v3->valueStr,"%d",ifIndex);
     v3->next = v4;

     sprintf(v4->varName.mibName,"mib2");
     sprintf(v4->varName.objName,"ifAdminStatus");
     sprintf(v4->varName.key,"%s",ifName);
     v4->varSyntax = integer;
     v4->valueStrLen = sprintf(v4->valueStr,"%d",ifAdminStatus);
     v4->next = v5;

     sprintf(v5->varName.mibName,"mib2");
     sprintf(v5->varName.objName,"ifOperStatus");
     sprintf(v5->varName.key,"%s",ifName);
     v5->varSyntax = integer;
     v5->valueStrLen = sprintf(v5->valueStr,"%d",ifOperStatus);
     v5->next = NULL;

     ret = xmpSendTrap(NULL,t,"mib2.linkUp",KORGOBJID,NULL);
     return ret;
}

/* ************************************************************* */
/* transition of ifOperStatus from some state to down state */
/* fix bug whereby we were losing variable-3 (sysObjectID) */

int sendLinkDownTrap(int ifIndex, int ifAdminStatus, int ifOperStatus, char *ifName)
{
     xmpPdu_t *t;
     xmpVariable_t *v1,*v2,*v3,*v4,*v5;
     int ret;

     fprintf(stderr,"sendLinkDownTrap: %d %d %d\n",ifIndex,ifAdminStatus,
             ifOperStatus);

     /* each of the variables needs to be malloc'd separately since
        they are each freed separately XXXX */

     /* alloc a PDU and get it set up */
     if ((t = (xmpPdu_t *)malloc(sizeof(xmpPdu_t))) == NULL) {
        return -1;
     }
     if ((v1 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return -1;
     }
     if ((v2 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return -1;
     }
     if ((v3 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return -1;
     }
     if ((v4 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return -1;
     }
     if ((v5 = (xmpVariable_t *)malloc(sizeof(xmpVariable_t))) == NULL) {
        free(t);
        return -1;
     }

     bzero((void *)t,sizeof(xmpPdu_t));
     bzero((void *)v1,sizeof(xmpVariable_t));
     bzero((void *)v2,sizeof(xmpVariable_t));
     bzero((void *)v3,sizeof(xmpVariable_t));
     bzero((void *)v4,sizeof(xmpVariable_t));
     bzero((void *)v5,sizeof(xmpVariable_t));

     t->msgType = Trap;
     t->messageID = (int64_t)time(NULL);
     t->errorStatus = noError;
     t->errorIndex = 0;
     t->numberVariables = 5;
     t->varList = v1;
     t->varListTail = v5;

     sprintf(v1->varName.mibName,"core");
     sprintf(v1->varName.objName,"trapType");
     v1->varSyntax = displayString;
     v1->valueStrLen = sprintf(v1->valueStr,"mib2.linkDown");
     v1->next = v2;

     sprintf(v2->varName.mibName,"core");
     sprintf(v2->varName.objName,"sysObjectID");
     v2->varSyntax = displayString;
     v2->valueStrLen = sprintf(v2->valueStr,KORGOBJID);
     v2->next = v3;

     sprintf(v3->varName.mibName,"mib2");
     sprintf(v3->varName.objName,"ifIndex");
     sprintf(v3->varName.key,"%s",ifName);
     v3->varSyntax = integer;
     v3->valueStrLen = sprintf(v3->valueStr,"%d",ifIndex);
     v3->next = v4;

     sprintf(v4->varName.mibName,"mib2");
     sprintf(v4->varName.objName,"ifAdminStatus");
     sprintf(v4->varName.key,"%s",ifName);
     v4->varSyntax = integer;
     v4->valueStrLen = sprintf(v4->valueStr,"%d",ifAdminStatus);
     v4->next = v5;

     sprintf(v5->varName.mibName,"mib2");
     sprintf(v5->varName.objName,"ifOperStatus");
     sprintf(v5->varName.key,"%s",ifName);
     v5->varSyntax = integer;
     v5->valueStrLen = sprintf(v5->valueStr,"%d",ifOperStatus);
     v5->next = NULL;

     ret = xmpSendTrap(NULL,t,"mib2.linkDown",KORGOBJID,NULL);
     return ret;
}

/* ************************************************************* */
