/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - mib2plugin.h */
/* rdk@krupczak.org */

/* cross-platform definitions for mib-2 plugins for various
   supported platforms (e.g. Linux, Solaris, NT)
*/

/* ************************************************************* */
/* platform-independent TCP connection structs */

typedef enum tcpState {
    tcpStateUndef = 0,
    tcpStateClosed = 1,
    tcpStateListen = 2,
    tcpStateSynSent = 3,
    tcpStateSynReceived = 4,
    tcpStateEstablished = 5,
    tcpStateFinWait1 = 6,
    tcpStateFinWait2= 7,
    tcpStateCloseWait = 8,
    tcpStateLastAck = 9,
    tcpStateClosing = 10,
    tcpStateTimeWait = 11,
    tcpStateDeleteTCB = 12,
} tcpState_t;

typedef enum udpState {
    udpStateUndef = 0,
    udpStateUnknown = 1,
    udpStateUnconnected = 2,
    udpStateConnected = 3, 
    udpStateUnbound = 4,
    udpStateIdle = 5,
} udpState_t;

typedef struct mib2TcpConn {

   u_int localIp;     /* addrs in network byte order; */
   u_int localPort;

   u_int remIp;        
   u_int remPort;   /* ports in host byte order */

   /* we make ports be a 4-byte int even though they
      technically can only be a two-byte entity
      since sscanf of %x requires a pointer to a
      4-byte integer */

   char localIPv6[48]; /* hex ascii chars */
   char remIPv6[48]; /* hex ascii chars */

   u_char ipVersion;  /* IPv4, IPv6 */

   /* others to be added */
   u_char state;      /* rfc1213 values */

   /* platform-specific data for mapping connections 
      back to processes */

   u_int64_t inode;     /* linux */
   pid_t pid;           /* win32 */
   
} mib2TcpConn_t;

typedef struct mib2UdpEndpt {

   u_int localIp;      /* addrs in network byte order */ 
   u_int localPort;

   u_int remIp;
   u_int remPort;   /* ports in host byte order */

   char localIPv6[48]; /* rfc2373 format string */
   char remIPv6[48]; /* rfc2373 format string */

   u_char ipVersion;  /* IPv4, IPv6 */

   /* other to be added */
   u_char state;  /* mib2 state */

   /* platform-specific data for mapping connections 
      back to processes */

   u_int64_t inode;     /* linux */
   pid_t pid;           /* win32 - Win2k, WinXP+ */
                        /* solaris unknown XXX */

} mib2UdpEndpt_t;

/* on linux, we can get raw sockets from /proc/net/raw 
   define the structure even though we (mib2) do not
   support them (yet) PROTO = IPPROTO_RAW */

typedef struct mib2RawEndpt {

   u_int localIp;      /* addrs in network byte order */ 
   u_int localPort;

   u_int remIp;
   u_int remPort;   /* ports in host byte order */

   char localIPv6[48]; /* rfc2373 format string */
   char remIPv6[48]; /* rfc2373 format string */

   u_char ipVersion;  /* IPv4, IPv6 */

   /* other to be added */
   u_char state;  /* mib2 state */

   /* platform-specific data for mapping connections 
      back to processes */

   u_int64_t inode;     /* linux */

} mib2RawEndpt_t;

typedef struct mib2IfEntry {

   char      ifname[256];  /* device name/description */
   char      alias[256];   /* alias like eth[0,1,2] lo[0,1,2] */

   /* receive stats */
   u_int64_t inbytes;     
   u_int64_t inpackets;   
   u_int64_t inerrors;    
   u_int64_t indrops;
   u_int64_t infifo;
   u_int64_t inframe;
   u_int64_t incompressed;
   u_int64_t inmulticast;

   u_int64_t outbytes;  /* transmit stats */
   u_int64_t outpackets;
   u_int64_t outerrors;
   u_int64_t outdrops;
   u_int64_t outfifo;
   u_int64_t outcollisions;
   u_int64_t outcarrier;
   u_int64_t outcompressed;

   u_int     mtu;
   u_char    hwaddr[12];
   u_int     hwaddrlen;

   int       ifOutQLen;
   u_int64_t ifOutErrors;
   u_int64_t ifOutDiscards;
   u_int64_t ifOutNUcastPkts;
   u_int64_t ifOutUcastPkts;
   u_int64_t ifOutOctets;
   u_int64_t ifInErrors;
   u_int64_t ifInOctets;
   u_int64_t ifInUnknownProtos;  
   u_int64_t ifInUcastPkts;
   u_int64_t ifInNUcastPkts;
   u_int64_t ifInDiscards;
   int       ifType;
   int       ifSpeed;
   int       ifAdminStatus;
   int       ifOperStatus;

   int       oldIfOperStatus;  /* for comparison */

   /* platform specific */
   u_char    ipVersion; /* IPv4 or IPv6 */

   u_int     flags;     /* device flags */ 
   u_int64_t lflags;    /* device flags (large) */ 
   u_int     ipv4addr;  /* first IP address for interface */
   u_int     ipv4bcast; /* broadcast addr */
   u_int     ipv4mask;  /* netmask for this interface */
   u_int     ifIndex;   /* win32 tracks this value */

   char      ipv6Addr[48]; /* rfc2373 format string */
   int       prefixLen;
   int       scope;

} mib2IfEntry_t;

typedef struct ipRoute {
   char ifname[256];   /* interface name */
   u_int ipv4dest;    /* dest ipv4 addr */
   u_int ipv4gway;    /* ipv4 gateway addr */
   u_int ipv4mask;    /* route mask for applying rule */

   u_int routeMetric; /* route metric */
   u_int routeType;   /* type of route */
   u_int routeProto;  /* how we got the route */
   u_int routeAge;    /* secs since update */

   /* platform specific flags/fields */
   u_int flags;
   u_int refcnt;
   u_int use;  
   u_int iretype; /* solaris, inet/ip.h IRE types */
   u_int ifIndex;  /* win32 tracks this */

   u_char ipVersion;
   char  ipv6dest[48]; /* rfc2373 format string */
   char  ipv6gway[48];
   u_int ipv6PrefixLen;

} ipRoute_t;

typedef struct arpEntry {
   char ifname[256];
   u_int ipv4addr;
   u_char hwaddr[12];
   int hwaddrlen;
   u_int hwtype;   /* RFC 826 HW type */

   /* platform specific XXX */   
   u_int flags;
   u_int ifIndex; /* win32 tracks this */

} arpEntry_t;

typedef struct ipAddrEntry {
 
   u_int  ipv4addr;
   char   ifname[256];
   u_int  ipv4mask;
   u_int  ipv4bcast;
   u_int  ipv4ReasmMaxSize;
   u_int  ifIndex;

   u_char ipVersion;  /* IPv4 or IPv6 */
   char   ipv6Addr[48]; /* rfc2373 format string */
   u_int  ipv6PrefixLen;
   u_int  ipv6AddrType;

} ipAddrEntry_t;

#define MIB2_CACHE_LIFE   30

/* ************************************************************* */
/* platform-independent stats structs */


/* we could have done an array of ints with the variable
   "hint" number being its index; it would have made
   the method routines a lot clearer but harder to read :) 
*/

typedef struct ipStats {
   int       ipForwarding;
   int       ipDefaultTTL;
   u_int64_t ipInReceives;
   u_int64_t ipInHdrErrors;
   u_int64_t ipInAddrErrors;
   u_int64_t ipForwDatagrams;
   u_int64_t ipInUnknownProtos;
   u_int64_t ipInDiscards;
   u_int64_t ipInDelivers;
   u_int64_t ipOutRequests;
   u_int64_t ipOutDiscards;
   u_int64_t ipOutNoRoutes;
   u_int64_t ipReasmTimeout;
   u_int64_t ipReasmReqds;
   u_int64_t ipReasmOKs;
   u_int64_t ipReasmFails;
   u_int64_t ipFragOKs;
   u_int64_t ipFragFails;
   u_int64_t ipFragCreates;
   u_int64_t ipRoutingDiscards;
} ipStats_t;

typedef struct icmpStats {
   u_int64_t icmpInMsgs;
   u_int64_t icmpInErrors;
   u_int64_t icmpInDestUnreachs;
   u_int64_t icmpInTimeExcds;
   u_int64_t icmpInParmProbs;
   u_int64_t icmpInSrcQuenchs;
   u_int64_t icmpInRedirects;
   u_int64_t icmpInEchos;
   u_int64_t icmpInEchoReps;
   u_int64_t icmpInTimestamps;
   u_int64_t icmpInTimestampReps;
   u_int64_t icmpInAddrMasks;
   u_int64_t icmpInAddrMaskReps;
   u_int64_t icmpOutMsgs;
   u_int64_t icmpOutErrors;
   u_int64_t icmpOutDestUnreachs;
   u_int64_t icmpOutTimeExcds;
   u_int64_t icmpOutParmProbs;
   u_int64_t icmpOutSrcQuenchs;
   u_int64_t icmpOutRedirects;
   u_int64_t icmpOutEchos;
   u_int64_t icmpOutEchoReps;
   u_int64_t icmpOutTimestamps;
   u_int64_t icmpOutTimestampReps;
   u_int64_t icmpOutAddrMasks;
   u_int64_t icmpOutAddrMaskReps;
} icmpStats_t;

typedef struct tcpStats {
    u_int64_t tcpRtoAlgorithm;
    u_int64_t tcpRtoMin;
    u_int64_t tcpRtoMax;
    int       tcpMaxConn;  /* could be negative */
    u_int64_t tcpActiveOpens;
    u_int64_t tcpPassiveOpens;
    u_int64_t tcpAttemptFails;
    u_int64_t tcpEstabResets;
    u_int64_t tcpCurrEstab;
    u_int64_t tcpInSegs;
    u_int64_t tcpOutSegs;
    u_int64_t tcpRetransSegs;
    u_int64_t tcpInErrs;
    u_int64_t tcpOutRsts;
} tcpStats_t;

typedef struct udpStats {
     u_int64_t udpInDatagrams;
     u_int64_t udpNoPorts;
     u_int64_t udpInErrors;
     u_int64_t udpOutDatagrams;
} udpStats_t;

/* ************************************************************* */
/* exported symbols, functions */

DLLExport int GetNumberMib2TcpConnections();
DLLExport int GetMib2TcpConnections(mib2TcpConn_t *, int);
DLLExport int GetNumberMib2Routes();
DLLExport int GetMib2Routes(ipRoute_t *, int);
DLLExport int GetNumberMib2Arps();
DLLExport int GetMib2ArpTable(arpEntry_t *, int);

int sendLinkDownTrap(int, int, int, char *);
int sendLinkUpTrap(int, int, int, char *);
int convertIPv6Addr(struct in6_addr *, char *, int);

/* ************************************************************* */
/* undocumented APIs and structures from iphlpapi.dll/lib/h */

#ifdef WIN32

#ifndef ANY_SIZE
#define ANY_SIZE 1
#endif

typedef struct {
  DWORD   dwState;        // state of the connection
  DWORD   dwLocalAddr;    // address on local computer
  DWORD   dwLocalPort;    // port number on local computer
  DWORD   dwRemoteAddr;   // address on remote computer
  DWORD   dwRemotePort;   // port number on remote computer
  DWORD	  dwProcessId;
} MIB_TCPEXROW, *PMIB_TCPEXROW;

typedef struct {
	DWORD			dwNumEntries;
	MIB_TCPEXROW	table[ANY_SIZE];
} MIB_TCPEXTABLE, *PMIB_TCPEXTABLE;

typedef struct {
  DWORD   dwLocalAddr;    // address on local computer
  DWORD   dwLocalPort;    // port number on local computer
  DWORD	  dwProcessId;
} MIB_UDPEXROW, *PMIB_UDPEXROW;

typedef struct {
	DWORD			dwNumEntries;
	MIB_UDPEXROW	table[ANY_SIZE];
} MIB_UDPEXTABLE, *PMIB_UDPEXTABLE;

#endif /* WIN32 */

/* ************************************************************* */
