/* 
   COPYRIGHT 2004 KRUPCZAK.ORG, LLC.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
   USA
 
   For more information, visit:
   http://www.krupczak.org/
*/

/* file - xmp_vars.c */
/* rdk@krupczak.org */

/* 
   functions for storing, registering, handling management variables
   for the agent and plugins 
*/

#include <pthread.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>

#ifndef WIN32
#include <strings.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/utsname.h>
#endif

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#include <windows.h>
#endif

#include <stdlib.h>

#include "xmp.h"
#include "globals.h"
#include "xmp_vars.h"
#include "xmp_session.h"
#include "rwlock.h"
#include "plugins.h"

/* 
   Apparently, on Linux, we need __USE_UNIX98 defined in order
   to get out types and definitions for rw-locks.  The rwlock
   functions are in the pthreads library, but symbol fun
   is preventing us from using it right now XXX 
   On solaris, reader/writer locks seem to be implemented fine
   according to the spec.  Sigh.  

   So, for now, use a mutex rather than a reader/write lock;
   Its less efficient but will suffice until we either implement
   our own reader/writer lock or we figure out how to mangle the
   symbols and compiler flags

   We found an implementation of reader/writer locks based
   on pthreads facilities so we use that rather than mutex
   Since its based on pthreads, it should be portable across
   all the UNIXs that support pthreads; we use this now.
*/

/* ************************************************************* */
/* doubly linked list of supported managed objects; first
   linked list of MIBs; for each MIB, linked list of objects
   within that MIB
*/

static xmpMib_t *xmpMibsHead = NULL;
static xmpMib_t *xmpMibsTail = NULL;

/* reader/writer lock for object list */
static rwlock_t varsLock = RWL_INITIALIZER;

/* ************************************************************* */
/* dont call xmpFindMIB if you already have a write lock!!!!! */

xmpMib_t *xmpFindMib(char *mibname)
{
     xmpMib_t *mptr;

     /* get read lock */
     rwl_readlock(&varsLock);

     mptr = xmpMibsHead;
     while (mptr != NULL) {
       if (strcmp(mptr->mibName,mibname) == 0) {
          break; /* out of while loop */
       }

       /* keep searching */
       mptr = mptr->next;
     }

     rwl_readunlock(&varsLock);

     return mptr;
}

/* ************************************************************* */
/* given a mibName, objName, return pointer to variable struct
   for now, we use a linear search over MIB object names
   and linear search over MIB names;
*/

xmpVar_t *xmpFindVar(char *mibname, char *objname)
{
     xmpVar_t *ptr;
     xmpMib_t *mptr;

     /* get read lock */
     rwl_readlock(&varsLock);

     mptr = xmpMibsHead;

     while (mptr != NULL) {
       if (strcmp(mibname,mptr->mibName) == 0)
 	  break;
       /* keep searching MIBs */
       mptr = mptr->next;
     }

     /* if we didnt find mib match, return now */
     if (mptr == NULL) {
        rwl_readunlock(&varsLock);
        return NULL;
     }

     /* now, given matching MIB, find matching variable */
     ptr = mptr->xmpVarsHead;

     while (ptr != NULL) {
       if (strcmp(ptr->objName,objname) == 0) {
	  break; /* out of while loop */
       }

       /* keep searching */
       ptr = ptr->next;
     }

     rwl_readunlock(&varsLock);

     return ptr;
}

/* ************************************************************* */
/* add a managed object to list */

DLLExport int xmpAddVar(xmpVar_t *var)
{
     xmpVar_t *ptr;
     xmpMib_t *mptr;
     int ret;
  
     if (var == NULL) {
        return -1;
     }

     if (xmpd_verbose) {
        printf("xmpAddVar: adding %s,%s\n",
	       var->mibName,var->objName);
     }

     /* check for duplicates -- write lock after checking for dups */
     ptr = xmpFindVar(var->mibName,var->objName);

     if (ptr != NULL) {
        printf("xmpAddVar: duplicate object %s,%s\n",
               var->mibName,var->objName);
        return 0;
     }

     /* get write lock */
     rwl_writelock(&varsLock);

     /* simple, initial case, no MIB */

     if (xmpd_verbose) {
        printf("xmpAddVar: new object, searching for matching MIB\n");
     }

     /* cant call xmpFindMib since we hold lock */
     mptr = xmpMibsHead;
     while (mptr != NULL) {
       if (strcmp(var->mibName,mptr->mibName) == 0) {
	  break;
       }
       mptr = mptr->next;
     }

     /* no matching MIB, create new one, connect, etc. */
     if (mptr == NULL) {

        if (xmpd_verbose) {
           printf("xmpAddVar: new MIB\n");
        }

        if ((mptr = malloc(sizeof(xmpMib_t))) == NULL) {
	   perror("malloc of mib struct failed");
           rwl_writeunlock(&varsLock);
           return -1;
        }
        bzero(mptr,sizeof(xmpMib_t)); /* !!! zero out */

        strncpy(mptr->mibName,var->mibName,127);

        mptr->xmpVarsHead = var;
        mptr->xmpVarsTail = var;
        mptr->next = mptr->prev = NULL; /* !!! zero out ptrs */

        var->next = NULL;
        var->prev = NULL;

        /* put new MIB struct into mib list */
        if (xmpMibsHead == NULL) {
           xmpMibsHead = mptr;
           xmpMibsTail = mptr;
           mptr->next = NULL;
           mptr->prev = NULL;
        }
        else {
 	   /* at least one entry on list, put new on tail */
	   mptr->prev = xmpMibsTail;
           if (mptr->prev)
	      mptr->prev->next = mptr;
	   xmpMibsTail = mptr;
        }
        rwl_writeunlock(&varsLock);
        return 1;

     } /* if (mptr == NULL), new MIB */

     /* mptr now points to desired MIB */

     if (xmpd_verbose) {
        printf("xmpAddVar: new object, existing MIB %s\n",mptr->mibName);
     }

     /* add alphabectically by MIB object name */
     ptr = mptr->xmpVarsHead;
     while (ptr != NULL) {
       ret = strcmp(var->objName,ptr->objName);
 
       if (ret < 0) {
 	  break; /* from while loop */
       }
       ptr = ptr->next;

     }

     if (ptr == NULL) {
       if (xmpd_verbose)
	  printf("xmpAddVar: adding %s to tail\n",var->objName);
        /* add to tail */
        var->prev = mptr->xmpVarsTail;
        var->next = NULL;
        mptr->xmpVarsTail->next = var;
        mptr->xmpVarsTail = var;
     }
     else {
        /* add before ptr since ptr is first object greater var's  */

        var->next = ptr;
        var->prev = ptr->prev;

        ptr->prev = var;
        /* ptr->next unchanged */

        if (var->prev != NULL)
 	   var->prev->next = var;

        if (mptr->xmpVarsHead == ptr)
	   mptr->xmpVarsHead = var;
     }

     rwl_writeunlock(&varsLock);
     return 1;
}

/* ************************************************************* */
/* dump vars for all MIBs */

int xmpDumpVars()
{
     xmpMib_t *mptr;
     xmpVar_t *ptr;

     /* get read lock */
     rwl_readlock(&varsLock);

     mptr = xmpMibsHead;

     while (mptr != NULL) {
       ptr = mptr->xmpVarsHead;

       while (ptr != NULL) {
         printf("MIB:%s Obj:%s Syntax:%d\n",ptr->mibName,ptr->objName,
                ptr->syntax); 

         ptr = ptr->next;

       } /* while objects within MIB */

       mptr = mptr->next;

     } /* while MIBs */


     rwl_readunlock(&varsLock);

     return 1;
}

/* for all MIBs, dump backwards */
int xmpDumpVarsBackwards()
{
     xmpVar_t *ptr;
     xmpMib_t *mptr;

     /* get read lock */
     rwl_readlock(&varsLock);


     mptr = xmpMibsTail;

     while (mptr != NULL) {

       ptr = mptr->xmpVarsTail;
 
       while (ptr != NULL) {
         printf("MIB-Backwards: %s Obj:%s\n",ptr->mibName, ptr->objName);
         ptr = ptr->prev;
       }

       mptr = mptr->prev;
     }

     rwl_readunlock(&varsLock);

     return 1;
}

/* ************************************************************* */
/* remove but do not free() the managed object */

DLLExport int xmpRemoveVarByName(char *mibname, char *objname)
{
     xmpVar_t *ptr;
     int ret;
     xmpMib_t *mptr;

     if (mibname == NULL) {
        return -1;
     }

     /* get write lock! */
     rwl_writelock(&varsLock);

     /* dont use xmpFindMib() since it tries to lock */

     /* find matching MIB; if no match, return */
     mptr = xmpMibsHead;
     while (mptr != NULL) {
       if (strcmp(mptr->mibName,mibname) == 0)
	  break;
       /* keep searching */
       mptr = mptr->next; 
     }
     if (mptr == NULL) {
        rwl_writeunlock(&varsLock);
        return -1;
     }
 
     /* null objname, remove everything and free MIB */
     if (objname == NULL) {

        /* this is really unnecessary; we simply have to unconnect
           or zero out the head and tail */

        while (mptr->xmpVarsHead != NULL) {
          mptr->xmpVarsHead = mptr->xmpVarsHead->next;
        }
        mptr->xmpVarsTail = NULL;

        if (mptr->prev != NULL)
           mptr->prev->next = mptr->next;

        if (mptr->next != NULL)
	   mptr->next->prev = mptr->prev;

        if (xmpMibsHead == mptr)
 	   xmpMibsHead = mptr->next;

        if (xmpMibsTail == mptr)
 	   xmpMibsTail = mptr->prev;

        free(mptr);

        rwl_writeunlock(&varsLock);

        return 1;
     }

     /* find MIB object matching target */

     ptr = mptr->xmpVarsHead;

     while (ptr != NULL) {
       ret = strcmp(ptr->mibName,mibname);
       
       if (ret == 0) {

 	  if ((objname != NULL) && (strcmp(ptr->objName,objname) != 0)) {
             ptr = ptr->next;
             continue; /* with while loop */
          }

          /* either we matched mib/object or just mib */

          if (ptr->prev != NULL)
 	     ptr->prev->next = ptr->next;

          if (ptr->next != NULL) 
 	     ptr->next->prev = ptr->prev;

          if (ptr == mptr->xmpVarsHead)
	     mptr->xmpVarsHead = ptr->next;

          if (ptr == mptr->xmpVarsTail)
	     mptr->xmpVarsTail = ptr->prev;
          
          ptr->next = ptr->prev = NULL;

       } /* if match on mibname */

       /* since objects are stored in alphabetical order by mib
          name, if the current ptr's mib name > than desired mib,
          we can break out of while loop and return */

       if (ret > 0) {
 	  break;
       }

       ptr = ptr->next;
     }

     rwl_writeunlock(&varsLock);

     return 1;
}

/* ************************************************************* */

xmpErrorStatus_t var_varObjects(char *key, xmpVariable_t *columns,
                                xmpPdu_t *out)
{
     xmpMib_t *mptr;
     xmpVar_t *ptr;
     char rowKey[256];
     xmpVariable_t *varptr, *cptr;
     u_int maxrows;
     int numRows = 0;
     int numVariables = 0;

     if ((key == NULL) || (columns == NULL) || (out == NULL)) {
        return genError;
     }

     if (xmpd_verbose)
        printf("var_varObjects: starting\n");

     /* keys can be MIBNAME.OBJNAME, partial, *, \0 */
     /* get the PDU ready */
     out->numberVariables = 0;
     out->varList = NULL;
     out->varListTail = NULL;
     maxrows = out->maxRows;
     if (maxrows <= 0)
         maxrows = (u_int)-1;

     /* get read lock */
     rwl_readlock(&varsLock);

     /* cycle through var list */
     mptr = xmpMibsHead;
     while (mptr != NULL) {

       ptr = mptr->xmpVarsHead;

       while (ptr != NULL) {

         /* construct key for this mib object */
         sprintf(rowKey,"%s.%s",ptr->mibName,ptr->objName);

         if ((strcmp(rowKey,key) != 0) && (key[0] != 0) && 
             (key[0] != '*')) {
	    ptr = ptr->next;
 	    continue; /* with while mib objects */
         }

         cptr = columns;
         while ((cptr != NULL) && (numRows < maxrows)) {

           /* we dont have to check if cptr->entryPtr valid since 
              we would not be called if there was not an internal
              entry for this column */

	   /* malloc a return value */
           if ((varptr = (xmpVariable_t *)malloc(sizeof(xmpVariable_t)))
               == NULL) {
	       perror("var_varObjects: malloc failed");
               out->errorStatus = genError;
               /* release lock */
               rwl_readunlock(&varsLock);
               return genError;
           }
           bcopy(cptr,varptr,sizeof(xmpVariable_t));
           strcpy(varptr->varName.key,rowKey);

           switch ( ((xmpVar_t *)cptr->entryPtr)->hint) {
	     case 1: /* varMibName */
	        varptr->varSyntax = displayString;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                             ptr->mibName);
                break;

  	     case 2: /* varObjname */
	        varptr->varSyntax = displayString;
                varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                             ptr->objName);
                break;

  	     case 3: /* varSyntax */
	        syntaxToString(ptr->syntax,varptr->valueStr,128);
                varptr->valueStrLen = strlen(varptr->valueStr);
                break;

  	     case 4: /* varTableName */
	        varptr->varSyntax = displayString;
                if (ptr->tableName[0] != 0)
                   varptr->valueStrLen = sprintf(varptr->valueStr,"%s",
                                                ptr->tableName);
                else {
		   varptr->valueStrLen = 0;
                   varptr->valueStr[0] = 0;
                }
                break;

	     default:
   	        fprintf(stderr,"var_varObjects: invalid column\n");
                /* release lock */
                rwl_readunlock(&varsLock);
                out->errorStatus = noSuchName;
                out->errorIndex = numVariables;
                out->numberVariables = numVariables;
                return noSuchName;

           } /* switch on column */

           /* glue into variable list */
           numVariables++;
           if (out->varListTail)
	      out->varListTail->next = varptr;
           out->varListTail = varptr;
           if (out->varList == NULL)
    	      out->varList = varptr;           

           cptr = cptr->next;
  
         } /* while columns */

         numRows++;         

         /* keep looping through vars */
         ptr = ptr->next;

       } /* while vars in MIB */

       /* keep looping through MIBs */
       mptr = mptr->next;

     } /* while MIBs */

     rwl_readunlock(&varsLock);

     out->numberVariables = numVariables;
     out->errorStatus = noError;
     if (numRows == out->maxRows)
        out->errorStatus = noErrorMoreData;

     return noError;
}

/* ************************************************************* */
/* do whatever we need to do to initialize variable structs, 
   etc. 
*/

int initVars()
{
     /* this function left intentionally blank :) */
     return 1;
}

/* ************************************************************* */

